
import ../types
import ./texture_decode
from dds_ktx import KtxInfo, KtxPart, get_ASTC_internal_format
import astc
export astc
import arr_ref
import stb_image_resize
import std/monotimes
import std/bitops

const myouEngineNumTextureThreads {.intdefine.} = 4
const myouEngineCompressTextures {.booldefine.} = true

const GL_COMPRESSED_RGBA_ASTC_12x12_KHR = 0x93BD'i32
const GL_COMPRESSED_SRGB8_ALPHA8_ASTC_12x12_KHR = 0x93DD'i32

template u32(x: untyped): uint32 = cast[uint32](x)

proc astc_compress_layer*(context: ptr AstcencContext, config: AstcencConfig,
           format: TextureFormat;  width, height: int, slices: seq[pointer]): ArrRef[byte] =
    let depth = slices.len
    let data_type = case format:
        of SRGB_u8, SRGB_Alpha_u8, R_u8, RG_u8, RGB_u8, RGBA_u8: ASTCENC_TYPE_U8
        of R_f16, RG_f16, RGB_f16, RGBA_f16: ASTCENC_TYPE_F16
        of R_f32, RG_f32, RGB_f32, RGBA_f32: ASTCENC_TYPE_F32
        else: raise ValueError.newException "Unsupported data type of " & $format
    let swizzle = case format.channel_count:
        # TODO: more optimal usage for 1/2 channels, normals, etc.
        of 1: AstcencSwizzle(r: SWZ_R, g: SWZ_R, b: SWZ_R, a: SWZ_1)
        of 2: AstcencSwizzle(r: SWZ_R, g: SWZ_G, b: SWZ_0, a: SWZ_1)
        of 3: AstcencSwizzle(r: SWZ_R, g: SWZ_G, b: SWZ_B, a: SWZ_1)
        else: AstcencSwizzle(r: SWZ_R, g: SWZ_G, b: SWZ_B, a: SWZ_A)
    let blocks_x = (width.u32 + config.block_x - 1) div config.block_x
    let blocks_y = (height.u32 + config.block_y - 1) div config.block_y
    let blocks_z = (depth.u32 + config.block_z - 1) div config.block_z
    let buffer_size = blocks_x * blocks_y * blocks_z * 16
    result = newArrRef[byte](buffer_size)
    let img = AstcencImage(
        dim_x: width.u32,
        dim_y: height.u32,
        dim_z: depth.u32,
        data_type: data_type,
        data: slices[0].addr,
    )
    const thread_index = 0 # TODO
    let err = astcenc_compress_image(context, img.addr, swizzle.addr,
        result[0].addr, buffer_size, thread_index)
    assert err == ASTCENC_SUCCESS, "ASTC encoding error: " & $err

proc atsc_compress*(tex: Texture, pixels: SliceMem[byte],
        callback: proc(self: Texture, info: KtxInfo, data: seq[KtxPart], refdata: seq[ArrRef[byte]])) =
    var width = tex.width.int32
    var height = tex.height.int32
    let depth = if tex.tex_type == TexCube: 1'i32 else: tex.depth.int32
    # TODO: when do we need ASTCENC_PRF_HDR?
    # maybe when A is connected to displacement?
    var profile: AstcencProfile
    var stbir_data_type: StbirDatatype
    case tex.format.component_type:
        of UByte:
            if tex.is_sRGB:
                profile = ASTCENC_PRF_LDR_SRGB
                stbir_data_type = STBIR_TYPE_UINT8_SRGB
            else:
                profile = ASTCENC_PRF_LDR
                stbir_data_type = STBIR_TYPE_UINT8
        of HalfFloat:
            profile = ASTCENC_PRF_HDR_RGB_LDR_A
            stbir_data_type = STBIR_TYPE_HALF_FLOAT
        of Float:
            profile = ASTCENC_PRF_HDR_RGB_LDR_A
            stbir_data_type = STBIR_TYPE_FLOAT
        of UShort:
            profile = ASTCENC_PRF_HDR_RGB_LDR_A
            stbir_data_type = STBIR_TYPE_UINT16
        else:
            assert false, "unreachable"
    # 4 because we're forcing 4 channels for the encoder
    let pixel_stride = 4 * (tex.format.stride div tex.format.channel_count)
    let quality = 10.0
    let flags = 0'u32
    var config: AstcencConfig
    let blk_size = (6,6)
    # TODO: 3D texture block size
    astc_assert astcenc_config_init(profile,
        blk_size[0].uint32, blk_size[1].uint32, 1'u32,
        quality, flags, config.addr)
    # config.progress_callback = proc(p: float32) {.cdecl.} =
    #     echo "progress ", p
    var ctx: ptr AstcencContext
    astc_assert astcenc_context_alloc(config.addr, 1, ctx)
    let time = getmonotime().ticks.float/1000000000
    var parts: seq[KtxPart]
    var data_refs: seq[ArrRef[byte]]
    var mip_level = 0'i32
    var w, prev_w = width
    var h, prev_h = height
    # round down to POT (TODO: make configurable)
    w = 1'i32 shl fastLog2(prev_w)
    h = 1'i32 shl fastLog2(prev_h)
    let resize_level_0 = w != prev_w or h != prev_h
    var current_level, last_level: ArrRef[byte]
    var last_src: pointer = pixels.data
    while (w > 4 and h > 4) or parts.len == 0:
        last_src = if mip_level != 0 or resize_level_0:
            current_level = newArrRef[byte](w*h*pixel_stride)
            stbir_resize(
                last_src, prev_w, prev_h, 0,
                current_level.toPointer, w, h, 0,
                StbirRgba, stbir_data_type,
                STBIR_EDGE_CLAMP, # TODO: handle other cases
                STBIR_FILTER_BOX,
            )
            last_level = current_level
            current_level.toPointer
        else:
            pixels.data
        let data = astc_compress_layer(ctx, config, tex.format, w, h, @[last_src])
        parts.add KtxPart(
            width: w, height: h, data: data.toPointer,
            len: data.byte_len, mip_level: mip_level,
        )
        data_refs.add data
        prev_w = w; prev_h = h
        w = max(1, w shr 1)
        h = max(1, h shr 1)
        mip_level.inc
    let time2 = getmonotime().ticks.float/1000000000
    echo "time: ", time2-time, " ", tex.name
    let info = KtxInfo(
        width: width, height: height, depth: depth,
        num_layers: 1, num_mipmaps: mip_level, has_alpha: tex.format.channel_count == 4,
        is_sRGB: tex.is_sRGB, is_astc: true,
        internal_format: get_ASTC_internal_format(blk_size, tex.is_sRGB)
    )
    callback(tex, info, parts, data_refs)
    astc_assert astcenc_compress_reset(ctx)
    astcenc_context_free(ctx)

type CallbackUncompressed = proc(tex: Texture, data: SliceMem[byte]) {.gcsafe.}
type CallbackCompressed = proc(tex: Texture, info: KtxInfo, data: seq[KtxPart], refdata: seq[ArrRef[byte]]) {.gcsafe.}

proc loadOptimized*(tex: Texture, slices: seq[SliceMem[byte]],
        callback_uncompressed: CallbackUncompressed = nil,
        callback_compressed: CallbackCompressed = nil,
        flip = true, min_channels = 0) {.gcsafe.} =
    
    var min_channels = min_channels
    var will_compress = myouEngineCompressTextures
    var will_load_uncompressed_first = false

    will_compress = will_compress and
        callback_compressed != nil and
        tex.format.component_type != UShort
    
    if will_compress:
        min_channels = 4

    tex.loadFileFromSlices(slices, proc(tex: Texture, pixels: SliceMem[byte]) {.gcsafe.} =
        if callback_uncompressed != nil and 
                (will_load_uncompressed_first or not will_compress):
            callback_uncompressed(tex, pixels)
        if will_compress:
            atsc_compress(tex, pixels, callback_compressed)
    , flip = flip, min_channels = min_channels)


# main -> thread  channels
type DecodeChanMsg = tuple[
    tex: Texture, slices: seq[SliceMem[byte]],
    callback_uncompressed: CallbackUncompressed,
    callback_compressed: CallbackCompressed,
    flip: bool, min_channels: int,
]
var decode_chan: Channel[DecodeChanMsg]
# main <- thread  channels
type DecodeReturnChanMsg = tuple[
    callback: CallbackUncompressed,
    tex: Texture, data: SliceMem[byte],
]
var decode_return_chan: Channel[DecodeReturnChanMsg]
type CompressedReturnChanMsg = tuple[
    callback: CallbackCompressed,
    tex: Texture, info: KtxInfo, data: seq[KtxPart], refdata: seq[ArrRef[byte]],
]
var compressed_return_chan: Channel[CompressedReturnChanMsg]

proc loadOptimizedThreaded*(tex: Texture, slices: seq[SliceMem[byte]],
        callback_uncompressed: CallbackUncompressed = nil,
        callback_compressed: CallbackCompressed = nil,
        flip = true, min_channels = 0) =
    
    decode_chan.send((tex: tex, slices: slices,
                    callback_uncompressed: callback_uncompressed,
                    callback_compressed: callback_compressed,
                    flip: flip, min_channels: min_channels))

var workers = newSeq[Thread[void]](myouEngineNumTextureThreads)
proc workerThreadProc() {.thread.} =
    # TODO: handle errors
    while true:
        let to_decode = decode_chan.recv()
        if to_decode.tex == nil:
            break
        proc cb(tex: Texture, data: SliceMem[byte]) =
            decode_return_chan.send((callback: to_decode.callback_uncompressed, tex: tex, data: data))
        proc cbc(tex: Texture, info: KtxInfo, data: seq[KtxPart], refdata: seq[ArrRef[byte]]) =
            compressed_return_chan.send((callback: to_decode.callback_compressed, tex: tex, 
            info: info, data: data, refdata: refdata))
        let cb_out = if to_decode.callback_uncompressed != nil: cb else: nil
        let cbc_out = if to_decode.callback_compressed != nil: cbc else: nil
        loadOptimized(to_decode.tex, to_decode.slices, cb_out, cbc_out,
                    to_decode.flip, to_decode.min_channels)

decode_chan.open()
decode_return_chan.open()
compressed_return_chan.open()
for worker in workers.mitems:
    worker.createThread(workerThreadProc)

proc updateTextureWorkerThreads*() =
    # TODO: handle errors
    while true:
        let tried = decode_return_chan.tryRecv()
        if not tried.dataAvailable:
            break
        let (cb, tex, data) = tried.msg
        cb(tex, data)
    while true:
        let tried = compressed_return_chan.tryRecv()
        if not tried.dataAvailable:
            break
        let (cb, tex, info, data, refdata) = tried.msg
        cb(tex, info, data, refdata)

proc terminateTextureWorkerThreads*() =
    for worker in workers:
        decode_chan.send(DecodeChanMsg.default)
    for worker in workers:
        worker.joinThread()
